﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.GrammaticalOptimization {
  public class ExhaustiveDepthFirstSearch {
    public event Action<string, double> FoundNewBestSolution;
    public event Action<string, double> SolutionEvaluated;

    private readonly int maxLen;
    private readonly Stack<string> stack = new Stack<string>();

    public ExhaustiveDepthFirstSearch(int maxLen) {
      this.maxLen = maxLen;      
    }

    public void Run(IProblem problem, int maxIterations) {
      double bestQuality = double.MinValue;
      stack.Push(problem.Grammar.SentenceSymbol.ToString());
      var sentences = GenerateLanguage(problem.Grammar);
      var sentenceEnumerator = sentences.GetEnumerator();
      for (int i = 0; sentenceEnumerator.MoveNext() && i < maxIterations; i++) {
        var sentence = sentenceEnumerator.Current;
        var quality = problem.Evaluate(sentence);
        RaiseSolutionEvaluated(sentence, quality);

        if (quality > bestQuality) {
          bestQuality = quality;
          RaiseFoundNewBestSolution(sentence, quality);
        }
      }
    }

    // create sentences lazily
    private IEnumerable<string> GenerateLanguage(IGrammar grammar) {
      while (stack.Any()) {
        var phrase = stack.Pop();

        var nt = phrase.First(grammar.IsNonTerminal);
        var ntIdx = phrase.IndexOf(nt); // TODO perf
        var alts = grammar.GetAlternatives(nt);
        foreach (var alt in alts) {
          var newPhrase = phrase.Remove(ntIdx, 1).Insert(ntIdx, alt);
          if (newPhrase.All(grammar.IsTerminal)) {
            yield return newPhrase;
          } else if (grammar.MinPhraseLength(newPhrase) <= maxLen) {
            stack.Push(newPhrase);
          }
        }
      }
    }


    private void RaiseSolutionEvaluated(string sentence, double quality) {
      var handler = SolutionEvaluated;
      if (handler != null) handler(sentence, quality);
    }
    private void RaiseFoundNewBestSolution(string sentence, double quality) {
      var handler = FoundNewBestSolution;
      if (handler != null) handler(sentence, quality);
    }
  }
}
